!+
! Module: createpseudoevt_mod
!
! Authors:  M.T. Ceballos -  SSC (IFCA) 
!           F.J. Carrera -  SSC (IFCA)  
!           X. Barcons - SSC (IFCA)
!          
! Description: Creates the pseudoEvt file with X/Y, DETX/DETY and THETA columns
!              from input image
!              
!

module createpseudoevt_mod

  use dal
  use cal
  use coordinate	
  use wcs_aux
  use constants   !DegToRad=pi/180
  use param  
  use errorHandling
  
  implicit none 
  
  private 

  public createpseudoevt

contains 

  !---------------------------------------------------------------------
  subroutine createpseudoevt
  !---------------------------------------------------------------------
    implicit none

    real(kind=single),dimension(:),allocatable :: xPixelsImg,yPixelsImg
    real(kind=single),dimension(:),allocatable :: xPixelsCol,yPixelsCol

    real(kind=double),dimension(:),allocatable :: tanX,tanY
    real(kind=double),dimension(:),allocatable :: dummyPhi,ra,dec
    real(kind=double), dimension(:), pointer :: coX_Data,coY_Data
    real(kind=double), dimension(:), pointer :: cotheta_Data
    real(kind=double) :: expStartTimeStamp

    integer(kind=Int32) :: i,j,ipix,irow,nrows,instrumentID
    integer(kind=Int32),dimension(:), pointer :: shape_inImage

    character(len=256) :: outEvt,inImage,inEvt,dateTimeStr
    
    type(DataSetT) :: set_outEvt,set_inEvt,set_inImage
    type(TableT):: tab_outEvt,tab_inEvt
    type(ArrayT) :: arr_inImage
    type(ColumnT) :: coX_outEvt,coY_outEvt,coTheta_outEvt
    type(SpacecraftAttitudeType) :: scAttitude
    type(WcsT):: wcsImg,wcsCol

!   read input parameters from param file
    
    outEvt=stringParameter("set")
    inImage=stringParameter("imageset")
    inEvt=stringParameter("inset")
    
    set_inImage=dataSet(inImage,Read)
    arr_inImage=array(set_inImage,"PRIMARY",Read)
    shape_inImage => dimensions(arr_inImage)
    set_inEvt=dataSet(inEvt,read)
    tab_inEvt=table(set_inEvt,"EVENTS",Read)
    nrows=shape_inImage(0)*shape_inImage(1)

!   Create columns in output event list: X,Y,THETA

    set_outEvt=dataSet(outEvt,CREATE)
    tab_outEvt=addtable(set_outEvt,"EVENTS",nrows)
    coX_outEvt=addcolumn(tab_outEvt,"X",Real64) 
    coY_outEvt=addcolumn(tab_outEvt,"Y",Real64)
    coTheta_outEvt=addcolumn(tab_outEvt,"THETA",Real64) 
    coX_Data=>Real64Data(coX_outEvt)
    coY_Data=>Real64Data(coY_outEvt)
    cotheta_Data=>real64Data(coTheta_outEvt)
                                  !arcs: angular deviation from optical axis



!
!   Reading WCS info in IMAGE and EVTLIST for coordinate transformation
!
    call readWcs(set_inImage, arr_inImage, wcsImg)
    call readWcs(tab_inEvt,"X","Y", wcsCol)


    if(stringAttribute(set_inImage,"INSTRUME") == "EMOS1")then
       instrumentID=4
    elseif(stringAttribute(set_inImage,"INSTRUME") == "EMOS2")then
       instrumentID=5
    elseif(stringAttribute(set_inImage,"INSTRUME") == "EPN")then
       instrumentID=6
    endif

    dateTimeStr=stringAttribute(set_inImage,"DATE-OBS")
    scAttitude%ra=real32Attribute(set_inImage,"RA_PNT")
    scAttitude%dec=real32Attribute(set_inImage,"DEC_PNT")
    scAttitude%apos=real32Attribute(set_inImage,"PA_PNT")

    expStartTimeStamp=toTimeTag(dateTimeStr)
    scAttitude%ra=scAttitude%ra*DegToRad
    scAttitude%dec=scAttitude%dec*DegToRad
    scAttitude%apos=scAttitude%apos*DegToRad

    call release(set_inImage)


!   Allocate arrays to store required coordinates

    allocate(xPixelsImg(0:nrows-1))
    allocate(yPixelsImg(0:nrows-1))
    allocate(xPixelsCol(0:nrows-1))
    allocate(yPixelsCol(0:nrows-1))
    allocate(tanX(0:nrows-1))
    allocate(tanY(0:nrows-1))
    allocate(ra(0:nrows-1))
    allocate(dec(0:nrows-1))


!   Populate initial arrays indexing image pixels 
    ipix=0
    do i=1,shape_inImage(0)
       do j=1,shape_inImage(1)
          xPixelsImg(ipix)=i
          yPixelsImg(ipix)=j
          ipix=ipix+1
       enddo
    enddo

!
!   Coordinate Transformation
!   -------------------------


!   Transforms pixels coordinates (i,j) in the image to tangent plane 
!   coordinates in rad using CRPIX1,CRPIX2,CDELT1,CDELT2 and CROTA2
    call pixelsToTan(xPixelsImg, yPixelsImg, wcsImg, tanX, tanY)
    deallocate(xPixelsImg,yPixelsImg)

!   From image tangent coordinates to RA,DEC using CRVAL1,CRVAL2
    call tanToRaDec(tanX,tanY,wcsImg%x%refWorld*DegToRad,&
         wcsImg%y%refWorld*DegToRad,ra,dec)

!   From image tangent coordinates to X,Y evtlist columns coordinates using
!   TCRPIX"X",TCRPIX"Y",TCDLT"X",TCDLT"Y" and CROTA2
    call tanToPixels(wcsCol,tanX,tanY,xPixelsCol,yPixelsCol)
    deallocate(tanX, tanY)


!   From RA,DEC coordinates to THETA polar coordinate
    call CAL_setState(instrument=instrumentId)
    allocate(dummyPhi(size(ra)))
    call raDecToTel(ra,dec,scAttitude,expStartTimeStamp,cotheta_Data,dummyPhi)
    deallocate(ra,dec,dummyPhi) 

!   Fill in columns
    do irow=0,nrows-1
       coX_Data(irow)=xPixelsCol(irow)
       coY_Data(irow)=yPixelsCol(irow)
    enddo
    deallocate(xPixelsCol,yPixelsCol)


    call release(set_outEvt);

  end subroutine createpseudoevt

end module createpseudoevt_mod
