!+
! Module: flmask_mod
!
! Authors:  M.T. Ceballos -  SSC (IFCA) 
!           F.J. Carrera -  SSC (IFCA)  
!           X. Barcons - SSC (IFCA)
!          
! Description: Creates the "positive" and "negative" mask images 
!              (rebinned pixels with counts above/below the median)
!              from input flatfielded image 
!
! Comments: Flatfielding carried out previously in Perl script
!           Mask images: to be used to extract fluctuations spectra
!                in Perl script
! 
module flmask_mod

  use dal
  use param  
  use errorHandling
  use dal_aux
  
  implicit none 
  
  private 

  public flmask

contains 

  !---------------------------------------------------------------------
  subroutine flmask
  !---------------------------------------------------------------------
    implicit none


    real(kind=single) :: inPixSize,binsize,att_val,tcrpxX,tcrpxY
    real(kind=double) :: median,median1,median2
    real(kind=double),dimension(:),allocatable :: binValid,arr_temp
    real(kind=double), dimension(:,:), pointer :: p_inImage
    real(kind=double), dimension(:,:), allocatable :: binImage,flatImage

    integer :: all_stat
    integer :: dim1_bin,dim2_bin,i,ib,j,jb,k!,isrc
    integer :: med,modulo,nofpix,valid
    integer :: pixInbin,type_inImage
    integer,dimension(:), pointer :: shape_inImage
    integer(kind=Int8),dimension(:,:), pointer :: p_posImage,p_negImage


    character(len=256) :: inEvt,inImage,negImage,posImage
    character(len=100) :: att_com !,srcRegfile
    character(len=10)  :: att_str
 
    logical :: alloc_ptr

    type(DataSetT) :: set_inEvt,set_inImage, set_posImage, set_negImage
    type(TableT):: tab_inEvt
    type(ArrayT) :: arr_posImage, arr_negImage, arr_inImage
    type(ColumnT) :: coX_inEvt, coY_inEvt

!   read input parameters from param file
    
    inEvt=stringParameter("set")
    inImage=stringParameter("imageset")
    binsize=realParameter("binsize")
    posImage=stringParameter("positivemaskset")
    negImage=stringParameter("negativemaskset")


    set_inEvt=dataSet(inEvt,Read)
    tab_inEvt=table(set_inEvt,"EVENTS")
    coX_inEvt=column(tab_inEvt,"X") 
    coY_inEvt=column(tab_inEvt,"Y") 
    tcrpxX=real32Attribute(coX_inEvt,"TCRPX")
    tcrpxY=real32Attribute(coY_inEvt,"TCRPX")
    call release(coX_inEvt)
    call release(coY_inEvt)
    call release(tab_inEvt)
    call release(set_inEvt)
    
    set_inImage=dataSet(inImage,Read)
    arr_inImage=array(set_inImage,"PRIMARY",Read)
    type_inImage=ArrayDataType(arr_inImage)
    shape_inImage => dimensions(arr_inImage)
    
    if(hasAttribute(set_inImage,"CDELT1"))then
       inPixsize=real32Attribute(set_inImage,"CDELT1")
    elseif(hasAttribute(arr_inImage,"CDELT1"))then
       inPixsize=real32Attribute(arr_inImage,"CDELT1")
    else
      write(errstr,*)"No keyword CDELT1 in:",inImage
      call fatal('NonStandardInput')
    endif
    inPixsize=ABS(inPixsize)   

!   keep a copy of Input Image in allocatable array flatImage to
!   make changes (subtract sources)

    allocate(flatImage(0:shape_inImage(0)-1,0:shape_inImage(1)-1))

    alloc_ptr=.TRUE.
    
    call readArrayData(inImage,p_inImage)
    flatImage=p_inImage;

!   create output "positive/negative" images (unbinned)

    set_posImage=dataSet(posImage,Create)
    arr_posImage=addArray(set_posImage,"IMAGE",Integer8,shape_inImage)
    p_posImage => int8Array2Data(arr_posImage)
    p_posImage=0

    set_negImage=dataSet(negImage,Create)
    arr_negImage=addArray(set_negImage,"IMAGE",Integer8,shape_inImage)
    p_negImage => int8Array2Data(arr_negImage)    
    p_negImage=0
 
!
!   copy WCS keywords from input image in masks
!
    att_val=real32Attribute(arr_inImage,"CRVAL1")
    att_com=label(arr_inImage,"CRVAL1")
    call setAttribute(arr_posImage,"CRVAL1",att_val,' ',trim(att_com))
    call setAttribute(arr_negImage,"CRVAL1",att_val,' ',trim(att_com))

    att_val=real32Attribute(arr_inImage,"CRVAL2")
    att_com=label(arr_inImage,"CRVAL2")
    call setAttribute(arr_posImage,"CRVAL2",att_val,' ',trim(att_com))
    call setAttribute(arr_negImage,"CRVAL2",att_val,' ',trim(att_com))
    
    att_val=real32Attribute(arr_inImage,"CRPIX1")
    att_com=label(arr_inImage,"CRPIX1")
    call setAttribute(arr_posImage,"CRPIX1",att_val,' ',trim(att_com))
    call setAttribute(arr_negImage,"CRPIX1",att_val,' ',trim(att_com))

    att_val=real32Attribute(arr_inImage,"CRPIX2")
    att_com=label(arr_inImage,"CRPIX2")
    call setAttribute(arr_posImage,"CRPIX2",att_val,' ',trim(att_com))
    call setAttribute(arr_negImage,"CRPIX2",att_val,' ',trim(att_com))

    att_val=real32Attribute(arr_inImage,"CDELT1")
    att_com=label(arr_inImage,"CDELT1")
    call setAttribute(arr_posImage,"CDELT1",att_val,' ',trim(att_com))
    call setAttribute(arr_negImage,"CDELT1",att_val,' ',trim(att_com))

    att_val=real32Attribute(arr_inImage,"CDELT2")
    att_com=label(arr_inImage,"CDELT2")
    call setAttribute(arr_posImage,"CDELT2",att_val,' ',trim(att_com))
    call setAttribute(arr_negImage,"CDELT2",att_val,' ',trim(att_com))

    att_str=stringAttribute(arr_inImage,"CTYPE1")
    att_com=label(arr_inImage,"CTYPE1")
    call setAttribute(arr_posImage,"CTYPE1",trim(att_str),trim(att_com))
    call setAttribute(arr_negImage,"CTYPE1",trim(att_str),trim(att_com))

    att_str=stringAttribute(arr_inImage,"CTYPE2")
    att_com=label(arr_inImage,"CTYPE2")
    call setAttribute(arr_posImage,"CTYPE2",trim(att_str),trim(att_com))
    call setAttribute(arr_negImage,"CTYPE2",trim(att_str),trim(att_com))
 
    call setAttribute(arr_posImage,"RADECSYS","FK5","World coord. system&
         & for this file")
    call setAttribute(arr_negImage,"RADECSYS","FK5","World coord. system&
         & for this file")

          
    if(alloc_ptr)deallocate(p_inImage)
    call release(arr_inImage)
    call release(set_inImage)



!   Create temporal array to keep the rebinned-corrected image

    if(binsize < inPixsize)then
       call fatal('BadParamValue','Requested parameter binsiz &
               & smaller than input image pixel size')
    endif

    pixInbin=FLOOR(binsize/inPixSize)

    dim1_bin=FLOOR(REAL(shape_inImage(0))/REAL(pixInbin))
    dim2_bin=FLOOR(REAL(shape_inImage(1))/REAL(pixInbin))

    if(dim1_bin==0 .OR. dim2_bin==0)then
       call fatal('BadParamValue','Requested parameter binsiz &
            & larger than input image size')
    elseif(dim1_bin==1 .AND. dim2_bin==1)then
       call fatal('BadParamValue','Requested parameter binsiz &
            & results in a binned image with only one pixel')
    endif


    allocate(binImage(0:dim1_bin-1,0:dim2_bin-1),stat=all_stat)
    if(all_stat /= 0)&
         call fatal('AllocationProblem','Failed Allocation for &
         & temporal Array binImage')


!   accumulate counts in each binned pixel of Binned Image

    binImage=0.d0
    valid=0
    binImageX: do ib=0,dim1_bin-1
       binImageY: do jb=0,dim2_bin-1
          nofpix=0
          flatImageX: do i=ib*pixInbin,(ib+1)*pixInbin-1
             flatImageY: do j=jb*pixInbin,(jb+1)*pixInbin-1
                if(flatImage(i,j) >= 0) then
                   nofpix=nofpix+1
                   binImage(ib,jb)=binImage(ib,jb)+ &
                        flatImage(i,j)
                endif
             enddo flatImageY
          enddo flatImageX
          
!         correct the counts in bin with the number of /=-1 pixels
          
          if (nofpix /= 0)then
             binImage(ib,jb)=binImage(ib,jb)*pixInbin**2./REAL(nofpix)
          else
             binImage(ib,jb)=-1.0d0
          endif
          
          if(binImage(ib,jb) >= 0)valid=valid+1
          
       enddo binImageY
    enddo binImageX


    if(valid==0)call fatal('NoPixLeft','Binned image does not have any pixel left after source extraction')
       
!   allocate memory for List of Pixels (>=0) from Binned Image 

    allocate(binValid(1:valid),stat=all_stat)
    if(all_stat /= 0)&
         call fatal('AllocationProblem','Failed Allocation for temporal &
         &array binValid')

    allocate(arr_temp(1:valid),stat=all_stat)
    if(all_stat /= 0)&
         call fatal('AllocationProblem','Failed Allocation for temporal &
         &array arr_temp')


!   Extract Binned-image array (pixels >=0) to a rank 1 array
        
    binValid=0.d0
    k=0
    do ib=0,dim1_bin-1 
       do jb=0,dim2_bin-1
          if(binImage(ib,jb) >= 0)then
             k=k+1
             binValid(k)=binImage(ib,jb)
          endif
       enddo
    enddo


!   calculate the median for pixels >= 0

    modulo=MOD(real(valid),2.0) 
    arr_temp=binValid    

    if(valid == 0)then
       
       call fatal('InvalidData','Invalid data in Input Image')
       
    elseif(modulo == 0)then
       
       med=valid/2
       median1=val_in_index(med,valid,arr_temp)

       med=valid/2 + 1
       arr_temp=binValid
       median2=val_in_index(med,valid,arr_temp)

       median=(median1+median2)/2.0

    elseif(modulo /= 0)then

       med=(valid+1)/2
       median=val_in_index(med,valid,arr_temp)
       
    endif

    deallocate(arr_temp, binValid)

!   fill in output "positive/negative" images (in input pixel size):
!      1 in binned pixels >= median  in "positive" Image
!      0 rest
!      1 in binned pixels < median  in "negative" Image
!      0 rest


    do ib=0,dim1_bin-1
       do jb=0,dim2_bin-1
          do i=ib*pixInbin,(ib+1)*pixInbin-1
             do j=jb*pixInbin,(jb+1)*pixInbin-1
                if(flatImage(i,j)>=0.and.binImage(ib,jb)>median &
                     .and.binImage(ib,jb)>=0)then
                   p_posImage(i,j)=1
                elseif(flatImage(i,j)>=0.and.binImage(ib,jb)<=median &
                     .and.binImage(ib,jb)>=0)then
                   p_negImage(i,j)=1
                endif
             enddo
          enddo
       enddo
    enddo
    

    deallocate(flatImage,stat=all_stat)
    if(all_stat /= 0)&
         call fatal('AllocationProblem','Failed Allocation for temporal &
         &array flatImage')

    deallocate(binImage,stat=all_stat)
    if(all_stat /= 0)&
         call fatal('AllocationProblem','Failed Allocation for temporal &
         &array binImage')

    call release(arr_posImage)
    call release(set_posImage)
    call release(arr_negImage)
    call release(set_negImage)

    
  end subroutine flmask


! -----------------------------------------------------------
! Function to calculate the value (val) of a n-dimensional "array"
! in position "index" 
! -----------------------------------------------------------
  function val_in_index(index, n, array) result (val)
    integer,intent(in) :: n,index    
    real(kind=double), dimension(n),intent(inout) :: array 
    integer :: i, ifin, j, ini, imed
    real(kind=double) :: a, temp, val
    
    ini = 1
    ifin = n
    
    do while((ifin-ini) > 1)
       imed = (ini + ifin) / 2
       temp = array(imed)
       array(imed) = array(ini + 1)
       array(ini + 1) = temp
       if (array(ini + 1) > array(ifin)) then
          temp = array(ini + 1)
          array(ini + 1) = array(ifin)
          array(ifin) = temp
       end if
       if (array(ini) > array(ifin)) then
          temp = array(ini)
          array(ini) = array(ifin)
          array(ifin) = temp
       end if
       if (array(ini + 1) > array(ini)) then
          temp = array(ini + 1)
          array(ini + 1) = array(ini)
          array(ini) = temp
       end if
       i = ini + 1
       j = ifin
       a = array(ini)
       do
          i = i + 1
          do while(array(i) < a)
             i=i+1
          enddo
          j = j - 1
          do while(array(j) > a)
             j = j - 1
          enddo
          
          if (j < i) then
             array(ini) = array(j)
             array(j) = a
             if (j >= index) ifin = j - 1
             if (j <= index) ini = i
             exit
          else
             temp = array(i)
             array(i) = array(j)
             array(j) = temp
          endif
       enddo
    enddo
    
    if ((ifin - ini) == 1) then
       if (array(ifin) < array(ini)) then
          temp = array(ini)
          array(ini) = array(ifin)
          array(ifin) = temp
       end if
    end if
    val = array(index)    
  end function val_in_index

  
end module flmask_mod  
  

